"""Classes for worksheets (.kicad_wks) and its contents

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    24.06.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List
from os import path

from kiutils.items.common import Justify
from kiutils.utils.strings import dequote
from kiutils.utils import sexpr
from kiutils.misc.config import KIUTILS_CREATE_NEW_GENERATOR_STR, KIUTILS_CREATE_NEW_VERSION_STR

@dataclass
class WksFontSize():
    """The ``WksFontSize`` token defines the size of a font in a worksheet"""

    width: float = 1.0
    """The ``width`` token defines the width of the font. Defaults to 1."""

    height: float = 1.0
    """The ``height`` token defines the height of the font. Defaults to 1."""

    @classmethod
    def from_sexpr(cls, exp: list) -> WksFontSize:
        """Convert the given S-Expresstion into a WksFontSize object

        Args:
            - exp (list): Part of parsed S-Expression ``(size ...)``

        Raises:
            - Exception: When given parameter's type is not a list or its length is not equal to 3
            - Exception: When the first item of the list is not ``size``

        Returns:
            - WksFontSize: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) != 3:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'size':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.width = exp[1]
        object.height = exp[2]
        return object

    def to_sexpr(self, indent=0, newline=False):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        return f'{indents}(size {self.width} {self.height}){endline}'

@dataclass
class WksFont():
    """The ``WksFont`` token defines how a text is drawn"""

    linewidth: Optional[float] = None
    """The optional ``linewidth`` token defines the width of the font's lines"""

    size: Optional[WksFontSize] = None
    """The optional ``size`` token defines the size of the font"""

    bold: bool = False
    """The ``bold`` token defines if the font is drawn bold. Defaults to False."""

    italic: bool = False
    """The ``italic`` token defines if the font is drawn italic. Defaults to False."""

    @classmethod
    def from_sexpr(cls, exp: list) -> WksFont:
        """Convert the given S-Expresstion into a WksFont object

        Args:
            - exp (list): Part of parsed S-Expression ``(font ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``font``

        Returns:
            - WksFont: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'font':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'bold': object.bold = True
                if item == 'italic': object.italic = True
                continue
            if item[0] == 'linewidth': object.linewidth = item[1]
            if item[0] == 'size': object.size = WksFontSize().from_sexpr(item)
        return object

    def to_sexpr(self, indent=0, newline=False):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object. Will return an empty string, if all members of this
                   class are set to ``None``.
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        lw = f' (linewidth {self.linewidth})' if self.linewidth is not None else ''
        size = f' {self.size.to_sexpr()}' if self.size is not None else ''
        bold = f' bold' if self.bold else ''
        italic = f' italic' if self.italic else ''

        if lw == '' and size == '' and bold == '' and italic == '':
            return ''
        else:
            return f'{indents}(font{lw}{size}{bold}{italic}){endline}'

@dataclass
class WksPosition():
    """The ``WksPosition`` token defines the positional coordinates and rotation of an worksheet
    object.
    """

    X: float = 0.0
    """The ``X`` attribute defines the horizontal position of the object. Defaults to 0."""

    Y: float = 0.0
    """The ``Y`` attribute defines the vertical position of the object. Defaults to 0."""

    corner: Optional[str] = None
    """The optional ``corner`` token is used to define the initial corner for repeating"""

    @classmethod
    def from_sexpr(cls, exp: list) -> WksPosition:
        """Convert the given S-Expresstion into a WksPosition object

        Args:
            - exp (list): Part of parsed S-Expression ``(xxx ...)``

        Raises:
            - Exception: When the given expression is not of type ``list`` or the list is less than
                         3 items long

        Returns:
            - WksPosition: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 3:
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.X = exp[1]
        object.Y = exp[2]

        # The last parameter refers to the corner token, if any is present
        if len(exp) > 3:
            object.corner = exp[3]

        return object

    def to_sexpr(self) -> str:
        """This object does not have a direct S-Expression representation."""
        raise NotImplementedError("This object does not have a direct S-Expression representation")

@dataclass
class Line():
    """The ``Line`` token defines how a line is drawn in a work sheet

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_graphical_line"""

    name: str = ""
    """The ``name`` token defines the name of the line object"""

    start: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``start`` token defines the start position of the line"""

    end: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``end`` token defines the end position of the line"""

    option: Optional[str] = None
    """The optional ``option`` token defines on which pages the line shall be shown. Possible values
    are:
    - None: Item will be shown on all pages
    - `notonpage1`: On all pages except page 1
    - `page1only`: Only visible on page 1"""

    lineWidth: Optional[float] = None
    """The optional ``lineWidth`` token attribute defines the width of the rectangle lines"""

    repeat: Optional[int] = None
    """The optional ``repeat`` token defines the count for repeated incremental lines"""

    incrx: Optional[float] = None
    """The optional ``incrx`` token defines the repeat distance on the X axis"""

    incry: Optional[float] = None
    """The optional ``incry`` token defines the repeat distance on the Y axis"""

    comment: Optional[str] = None
    """The optional ``comment`` token is a comment for the line object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Line:
        """Convert the given S-Expresstion into a Line object

        Args:
            - exp (list): Part of parsed S-Expression ``(line ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``tbtext``

        Returns:
            - Line: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'line':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'start': object.start = WksPosition().from_sexpr(item)
            if item[0] == 'end': object.end = WksPosition().from_sexpr(item)
            if item[0] == 'option': object.option = item[1]
            if item[0] == 'linewidth': object.lineWidth = item[1]
            if item[0] == 'repeat': object.repeat = item[1]
            if item[0] == 'incrx': object.incrx = item[1]
            if item[0] == 'incry': object.incry = item[1]
            if item[0] == 'comment': object.comment = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        start_corner = f' {self.start.corner}' if self.start.corner is not None else ''
        end_corner = f' {self.end.corner}' if self.end.corner is not None else ''
        option = f' (option {self.option})' if self.option is not None else ''
        repeat = f' (repeat {self.repeat})' if self.repeat is not None else ''
        incrx = f' (incrx {self.incrx})' if self.incrx is not None else ''
        incry = f' (incry {self.incry})' if self.incry is not None else ''
        comment = f' (comment "{dequote(self.comment)}")\n' if self.comment is not None else ''
        lw = f' (linewidth {self.lineWidth})' if self.lineWidth is not None else ''

        expression  = f'{indents}(line (name "{dequote(self.name)}") '
        expression += f'(start {self.start.X} {self.start.Y}{start_corner}) '
        expression += f'(end {self.end.X} {self.end.Y}{end_corner})'
        expression += f'{option}{lw}{repeat}{incrx}{incry}{comment}){endline}'
        return expression

@dataclass
class Rect():
    """The ``Rect`` token defines how a rectangle is drawn in a work sheet

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_graphical_rectangle"""

    name: str = ""
    """The ``name`` token defines the name of the rectangle object"""

    start: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``start`` token defines the start position of the rectangle"""

    end: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``end`` token defines the end position of the rectangle"""

    option: Optional[str] = None
    """The optional ``option`` token defines on which pages the rectangle shall be shown. Possible values
    are:
    - None: Item will be shown on all pages
    - `notonpage1`: On all pages except page 1
    - `page1only`: Only visible on page 1"""

    lineWidth: Optional[float] = None
    """The optional ``lineWidth`` token attribute defines the width of the rectangle lines"""

    repeat: Optional[int] = None
    """The optional ``repeat`` token defines the count for repeated incremental rectangles"""

    incrx: Optional[float] = None
    """The optional ``incrx`` token defines the repeat distance on the X axis"""

    incry: Optional[float] = None
    """The optional ``incry`` token defines the repeat distance on the Y axis"""

    comment: Optional[str] = None
    """The optional ``comment`` token is a comment for the rectangle object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Rect:
        """Convert the given S-Expresstion into a Rect object

        Args:
            - exp (list): Part of parsed S-Expression ``(rect ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``rect``

        Returns:
            - Rect: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'rect':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'start': object.start = WksPosition().from_sexpr(item)
            if item[0] == 'end': object.end = WksPosition().from_sexpr(item)
            if item[0] == 'option': object.option = item[1]
            if item[0] == 'linewidth': object.lineWidth = item[1]
            if item[0] == 'repeat': object.repeat = item[1]
            if item[0] == 'incrx': object.incrx = item[1]
            if item[0] == 'incry': object.incry = item[1]
            if item[0] == 'comment': object.comment = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        start_corner = f' {self.start.corner}' if self.start.corner is not None else ''
        end_corner = f' {self.end.corner}' if self.end.corner is not None else ''
        option = f' (option {self.option})' if self.option is not None else ''
        repeat = f' (repeat {self.repeat})' if self.repeat is not None else ''
        incrx = f' (incrx {self.incrx})' if self.incrx is not None else ''
        incry = f' (incry {self.incry})' if self.incry is not None else ''
        comment = f' (comment "{dequote(self.comment)}")\n' if self.comment is not None else ''
        lw = f' (linewidth {self.lineWidth})' if self.lineWidth is not None else ''

        expression  = f'{indents}(rect (name "{dequote(self.name)}") '
        expression += f'(start {self.start.X} {self.start.Y}{start_corner}) '
        expression += f'(end {self.end.X} {self.end.Y}{end_corner})'
        expression += f'{option}{lw}{repeat}{incrx}{incry}{comment}){endline}'
        return expression

@dataclass
class Polygon():
    """The ``Polygon`` token defines a graphical polygon in a worksheet

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_graphical_polygon
    """

    name: str = ""
    """The ``name`` token defines the name of the polygon"""

    position: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``position`` token defines the coordinates of the polygon"""

    option: Optional[str] = None
    """The optional ``option`` token defines on which pages the polygon shall be shown. Possible values
    are:
    - None: Item will be shown on all pages
    - `notonpage1`: On all pages except page 1
    - `page1only`: Only visible on page 1"""

    rotate: Optional[float] = None
    """The optional ``rotate`` token defines the rotation angle of the polygon object"""

    coordinates: List[WksPosition] = field(default_factory=list)
    """The ``coordinates`` token defines a list of X/Y coordinates that forms the polygon"""

    repeat: Optional[int] = None
    """The optional ``repeat`` token defines the count for repeated incremental polygons"""

    incrx: Optional[float] = None
    """The optional ``incrx`` token defines the repeat distance on the X axis"""

    incry: Optional[float] = None
    """The optional ``incry`` token defines the repeat distance on the Y axis"""

    comment: Optional[str] = None
    """The optional ``comment`` token is a comment for the polygon object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Polygon:
        """Convert the given S-Expresstion into a Polygon object

        Args:
            - exp (list): Part of parsed S-Expression ``(polygon ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``polygon``

        Returns:
            - Polygon: Object of the class initialized with the given S-Expression
        """
        # TODO: Polygons seem to not be available in the WKS editor GUI. Are those still a feature?
        raise NotImplementedError("Polygons are not yet handled! Please report this bug along with the file being parsed.")

    def to_sexpr(self, indent=0, newline=False):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        raise NotImplementedError("Polygons are not yet handled! Please report this bug along with the file being parsed.")

@dataclass
class Bitmap():
    """The ``Polygon`` token defines on or more embedded images

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_image
    """

    name: str = ""
    """The ``name`` token defines the name of the bitmap"""

    position: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``position`` token defines the coordinates of the bitmap"""

    option: Optional[str] = None
    """The optional ``option`` token defines on which pages the image shall be shown. Possible values
    are:
    - None: Item will be shown on all pages
    - `notonpage1`: On all pages except page 1
    - `page1only`: Only visible on page 1"""

    scale: float = 1.0
    """The ``scale`` token defines the scale of the bitmap object"""

    repeat: Optional[int] = None
    """The optional ``repeat`` token defines the count for repeated incremental bitmaps"""

    incrx: Optional[float] = None
    """The optional ``incrx`` token defines the repeat distance on the X axis"""

    incry: Optional[float] = None
    """The optional ``incry`` token defines the repeat distance on the Y axis"""

    # Comments seem to be buggy as of 25.06.2022 ..
    comment: Optional[str] = None
    """The optional ``comment`` token is a comment for the bitmap object"""

    # TODO: Parse this nonesense as a binary struct to make it more useful
    pngdata: List[str] = field(default_factory=list)
    """The ``pngdata`` token defines a list of strings representing up to 32 bytes per entry of
    the image being saved.

    Format:
    - "xx xx xx xx xx (..) xx "

    The list must be 32byte aligned, leaving a space after the last byte as shown in the format
    example.
    """

    @classmethod
    def from_sexpr(cls, exp: list) -> Bitmap:
        """Convert the given S-Expresstion into a Bitmap object

        Args:
            - exp (list): Part of parsed S-Expression ``(bitmap ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``bitmap``

        Returns:
            - Bitmap: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'bitmap':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'pos': object.position = WksPosition().from_sexpr(item)
            if item[0] == 'option': object.option = item[1]
            if item[0] == 'scale': object.scale = item[1]
            if item[0] == 'repeat': object.repeat = item[1]
            if item[0] == 'incrx': object.incrx = item[1]
            if item[0] == 'incry': object.incry = item[1]
            if item[0] == 'comment': object.comment = item[1]
            if item[0] == 'pngdata':
                if len(item) < 2: continue
                for data in item[1:]:
                    if data[0] != 'data': continue
                    object.pngdata.append(data[1])
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        repeat = f' (repeat {self.repeat})' if self.repeat is not None else ''
        incrx = f' (incrx {self.incrx})' if self.incrx is not None else ''
        incry = f' (incry {self.incry})' if self.incry is not None else ''
        option = f' (option {self.option})' if self.option is not None else ''
        corner = f' {self.position.corner}' if self.position.corner is not None else ''

        expression  = f'{indents}(bitmap (name "{dequote(self.name)}") '
        expression += f'(pos {self.position.X} {self.position.Y}{corner}){option} (scale {self.scale})'
        expression += f'{repeat}{incrx}{incry}\n'
        if self.comment is not None:
            # Here KiCad decides to only use 1 space for some unknown reason ..
            expression += f' (comment "{dequote(self.comment)}")\n'
        expression += f'{indents}(pngdata\n'
        for data in self.pngdata:
            expression += f'{indents}  (data "{data}")\n'
        expression += f'{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression


@dataclass
class TbText():
    """The ``TbText`` token define text used in the title block of a work sheet

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_title_block_text"""

    text: str = ""
    """The ``text`` token defines the text itself"""

    name: str = ""
    """The ``name`` token defines the name of the text object"""

    position: WksPosition = field(default_factory=lambda: WksPosition())
    """The ``position`` token defines the position of the text"""

    option: Optional[str] = None
    """The optional ``option`` token defines on which pages the text shall be shown. Possible values
    are:
    - None: Item will be shown on all pages
    - `notonpage1`: On all pages except page 1
    - `page1only`: Only visible on page 1"""

    rotate: Optional[float] = None
    """The optional ``rotate`` token defines the rotation of the text in degrees"""

    font: WksFont = field(default_factory=lambda: WksFont())
    """The ``font`` token define how the text is drawn"""

    justify: Optional[Justify] = None
    """The optional ``justify`` token defines the justification of the text"""

    maxlen: Optional[float] = None
    """The optional ``maxlen`` token defines the maximum length of the text"""

    maxheight: Optional[float] = None
    """The optional ``maxheight`` token defines the maximum height of the text"""

    repeat: Optional[int] = None
    """The optional ``repeat`` token defines the count for repeated incremental text"""

    incrx: Optional[float] = None
    """The optional ``incrx`` token defines the repeat distance on the X axis"""

    incry: Optional[float] = None
    """The optional ``incry`` token defines the repeat distance on the Y axis"""

    incrlabel: Optional[int] = None
    """The optional ``incrlabel`` token defines the amount of characters that are moved with every
    repeated incremental text"""

    comment: Optional[str] = None
    """The optional ``comment`` token is a comment for the text object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> TbText:
        """Convert the given S-Expresstion into a TbText object

        Args:
            - exp (list): Part of parsed S-Expression ``(tbtext ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``tbtext``

        Returns:
            - TbText: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'tbtext':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'pos': object.position = WksPosition().from_sexpr(item)
            if item[0] == 'option': object.option = item[1]
            if item[0] == 'rotate': object.rotate = item[1]
            if item[0] == 'font': object.font = WksFont().from_sexpr(item)
            if item[0] == 'justify': object.justify = Justify().from_sexpr(item)
            if item[0] == 'maxlen': object.maxlen = item[1]
            if item[0] == 'maxheight': object.maxheight = item[1]
            if item[0] == 'repeat': object.repeat = item[1]
            if item[0] == 'incrx': object.incrx = item[1]
            if item[0] == 'incry': object.incry = item[1]
            if item[0] == 'incrlabel': object.incrlabel = item[1]
            if item[0] == 'comment': object.comment = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        corner = f' {self.position.corner}' if self.position.corner is not None else ''
        repeat = f' (repeat {self.repeat})' if self.repeat is not None else ''
        incrx = f' (incrx {self.incrx})' if self.incrx is not None else ''
        incry = f' (incry {self.incry})' if self.incry is not None else ''
        option = f' (option {self.option})' if self.option is not None else ''
        rotate = f' (rotate {self.rotate})' if self.rotate is not None else ''
        justify = f' {self.justify.to_sexpr()}' if self.justify is not None else ''
        maxlen = f' (maxlen {self.maxlen})' if self.maxlen is not None else ''
        maxheight = f' (maxheight {self.maxheight})' if self.maxheight is not None else ''
        incrlabel = f' (incrlabel {self.incrlabel})' if self.incrlabel is not None else ''
        font = f' {self.font.to_sexpr()}' if self.font.to_sexpr() != '' else ''

        expression  = f'{indents}(tbtext "{dequote(self.text)}" (name "{dequote(self.name)}") '
        expression += f'(pos {self.position.X} {self.position.Y}{corner}){option}{rotate}'
        expression += f'{font}{justify}{maxlen}{maxheight}{repeat}{incrx}{incry}{incrlabel}'
        if self.comment is not None:
            expression += f' (comment "{dequote(self.comment)}")\n'
        expression += f'){endline}'
        return expression


@dataclass
class TextSize():
    """The ``TextSize`` define the default width and height of text"""

    width: float = 1.5
    """The ``width`` token defines the default width of a text element. Defaults to 1,5."""

    height: float = 1.5
    """The ``height`` token defines the default height of a text element. Defaults to 1,5."""

    @classmethod
    def from_sexpr(cls, exp: list) -> TextSize:
        """Convert the given S-Expresstion into a TextSize object

        Args:
            - exp (list): Part of parsed S-Expression ``(textsize ...)``

        Raises:
            - Exception: When given parameter's type is not a list or when its not exactly 3 long
            - Exception: When the first item of the list is not ``textsize``

        Returns:
            - TextSize: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) != 3:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'textsize':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.width = exp[1]
        object.height = exp[2]
        return object

    def to_sexpr(self, indent=0, newline=False):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        return f'{indents}(textsize {self.width} {self.height}){endline}'

@dataclass
class Setup():
    """The ``setup`` token defines the configuration information for the work sheet

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_set_up_section"""

    textSize: TextSize = field(default_factory=lambda: TextSize())
    """The ``textSize`` token defines the default width and height of text"""

    lineWidth: float = 0.15
    """The ``lineWidth`` token attribute defines the default width of lines. Defaults to 0,15."""

    textLineWidth: float = 0.15
    """The ``textLineWidth`` token attribute define the default width of the lines used to draw 
    text. Defaults to 0,15."""

    leftMargin: float = 10.0
    """The ``leftMargin`` token defines the distance from the left edge of the page"""

    rightMargin: float = 10.0
    """The ``rightMargin`` token defines the distance from the right edge of the page"""

    topMargin: float = 10.0
    """The ``topMargin`` token defines the distance from the top edge of the page"""

    bottomMargin: float = 10.0
    """The ``bottomMargin`` token defines the distance from the bottom edge of the page"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Setup:
        """Convert the given S-Expresstion into a Setup object

        Args:
            - exp (list): Part of parsed S-Expression ``(setup ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``setup``

        Returns:
            - Setup: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'setup':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'textsize': object.textSize = TextSize().from_sexpr(item)
            if item[0] == 'linewidth': object.lineWidth = item[1]
            if item[0] == 'textlinewidth': object.textLineWidth = item[1]
            if item[0] == 'left_margin': object.leftMargin = item[1]
            if item[0] == 'right_margin': object.rightMargin = item[1]
            if item[0] == 'top_margin': object.topMargin = item[1]
            if item[0] == 'bottom_margin': object.bottomMargin = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        # KiCad puts no spaces between tokens here
        expression =  f'{indents}(setup {self.textSize.to_sexpr()}(linewidth {self.lineWidth})'
        expression += f'(textlinewidth {self.textLineWidth})\n{indents}'
        expression += f'(left_margin {self.leftMargin})(right_margin {self.rightMargin})'
        expression += f'(top_margin {self.topMargin})(bottom_margin {self.bottomMargin})'
        expression += f'){endline}'

        return expression

@dataclass
class WorkSheet():
    """The ``WorkSheet`` token defines a KiCad worksheet (.kicad_wks file)

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-worksheet/#_header_section"""

    version: str = KIUTILS_CREATE_NEW_VERSION_STR
    """The ``version`` token defines the work sheet version using the YYYYMMDD date format"""

    generator: str = KIUTILS_CREATE_NEW_GENERATOR_STR
    """The ``generator`` token defines the program used to write the file"""

    setup: Setup = field(default_factory=lambda: Setup())
    """The ``setup`` token defines the configuration information for the work sheet"""

    drawingObjects: List = field(default_factory=list)
    """The ``drawingObjects`` token can contain zero or more texts, lines, rectangles, polys or images"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the board file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> WorkSheet:
        """Convert the given S-Expresstion into a WorkSheet object

        Args:
            - exp (list): Part of parsed S-Expression ``(kicad_wks ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``kicad_wks``

        Returns:
            - WorkSheet: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'kicad_wks':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'version': object.version = item[1]
            if item[0] == 'generator': object.generator = item[1]
            if item[0] == 'setup': object.setup = Setup().from_sexpr(item)
            if item[0] == 'rect': object.drawingObjects.append(Rect().from_sexpr(item))
            if item[0] == 'line': object.drawingObjects.append(Line().from_sexpr(item))
            if item[0] == 'polygon': object.drawingObjects.append(Polygon().from_sexpr(item))
            if item[0] == 'tbtext': object.drawingObjects.append(TbText().from_sexpr(item))
            if item[0] == 'bitmap': object.drawingObjects.append(Bitmap().from_sexpr(item))
        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> WorkSheet:
        """Load a worksheet directly from a KiCad worksheet file (`.kicad_wks`) and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - WorkSheet: Object of the WorkSheet class initialized with the given KiCad worksheet
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            item = cls.from_sexpr(sexpr.parse_sexp(infile.read()))
            item.filePath = filepath
            return item

    @classmethod
    def create_new(cls) -> WorkSheet:
        """Creates a new empty worksheet as KiCad would create it

        Returns:
            WorkSheet: A empty worksheet
        """
        return cls(
            version = KIUTILS_CREATE_NEW_VERSION_STR,
            generator = KIUTILS_CREATE_NEW_GENERATOR_STR
        )

    def to_file(self, filepath = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set, 
                                        the attribute ``self.filePath`` will be used instead.

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w') as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(kicad_wks (version {self.version}) (generator {self.generator})\n'
        expression += self.setup.to_sexpr(indent+2)
        for item in self.drawingObjects:
            expression += item.to_sexpr(indent+2)
        expression += f'{indents}){endline}'

        return expression


"""Configuration variables

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    19.09.2022 - created
"""

KIUTILS_CREATE_NEW_VERSION_STR = '20211014'
"""Version string used in ``create_new()`` class functions"""

KIUTILS_CREATE_NEW_GENERATOR_STR = 'kiutils'
"""Generator string used in ``create_new()`` class functions"""


"""Class to manage KiCad schematics

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    19.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List, Union
from os import path

from kiutils.items.common import Image, PageSettings, TitleBlock
from kiutils.items.schitems import *
from kiutils.symbol import Symbol
from kiutils.utils import sexpr
from kiutils.misc.config import KIUTILS_CREATE_NEW_GENERATOR_STR, KIUTILS_CREATE_NEW_VERSION_STR

@dataclass
class Schematic():
    """The ``schematic`` token represents a KiCad schematic as defined by the schematic file format

    Documenatation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/
    """

    version: str = KIUTILS_CREATE_NEW_VERSION_STR
    """The ``version`` token attribute defines the schematic version using the YYYYMMDD date format"""

    generator: str = KIUTILS_CREATE_NEW_GENERATOR_STR
    """The ``generator`` token attribute defines the program used to write the file"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    paper: PageSettings = field(default_factory=lambda: PageSettings())
    """The ``paper`` token defines the drawing page size and orientation"""

    titleBlock: Optional[TitleBlock] = None
    """The ``titleBlock`` token defines author, date, revision, company and comments of the schematic"""

    libSymbols: List[Symbol] = field(default_factory=list)
    """The ``libSymbols`` token defines a list of symbols that are used in the schematic"""

    schematicSymbols: List[SchematicSymbol] = field(default_factory=list)
    """The ``schematicSymbols`` token defines a list of instances of symbols used in the schematic"""

    junctions: List[Junction] = field(default_factory=list)
    """The ``junctions`` token defines a list of junctions used in the schematic"""

    noConnects: List[NoConnect] = field(default_factory=list)
    """The ``noConnect`` token defines a list of no_connect markers used in the schematic"""

    busEntries: List[BusEntry] = field(default_factory=list)
    """The ``busEntries`` token defines a list of bus_entry used in the schematic"""

    graphicalItems: List[Union[Connection, PolyLine]] = field(default_factory=list)
    """The ``graphicalItems`` token defines a list of ``bus``, ``wire`` or ``polyline`` elements 
    used in the schematic"""

    shapes: List[Union[Arc, Circle, Rectangle]] = field(default_factory=list)
    """The ``shapes`` token defines a list of graphical shapes (``Arc``, ``Rectangle`` or 
    ``Circle``) used in the schematic.
    
    Available since KiCad v7"""

    images: List[Image] = field(default_factory=list)
    """The ``images`` token defines a list of images used in the schematic"""

    texts: List[Text] = field(default_factory=list)
    """The ``text`` token defines a list of texts used in the schematic"""

    textBoxes: List[TextBox] = field(default_factory=list)
    """The ``text_box`` token defines a list of text boxes used in the schematic"""

    labels: List[LocalLabel] = field(default_factory=list)
    """The ``labels`` token defines a list of local labels used in the schematic"""

    globalLabels: List[GlobalLabel] = field(default_factory=list)
    """The ``globalLabels`` token defines a list of global labels used in the schematic"""

    hierarchicalLabels: List[HierarchicalLabel] = field(default_factory=list)
    """The ``herarchicalLabels`` token defines a list of hierarchical labels used in the schematic"""

    netclassFlags: List[NetclassFlag] = field(default_factory=list)
    """The ``netclassFlags`` token defines a list of netclass flags used in the schematic.
    
    Available since KiCad v7"""

    sheets: List[HierarchicalSheet] = field(default_factory=list)
    """The ``sheets`` token defines a list of hierarchical sheets used in the schematic"""

    sheetInstances: List[HierarchicalSheetInstance] = field(default_factory=list)
    """The ``sheetInstances`` token defines a list of instances of hierarchical sheets used in
    the schematic"""

    symbolInstances: List[SymbolInstance] = field(default_factory=list)
    """The ``symbolInstances`` token defines a list of instances of symbols from ``libSymbols`` token
    used in the schematic"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the schematic file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Schematic:
        """Convert the given S-Expresstion into a Schematic object

        Args:
            - exp (list): Part of parsed S-Expression ``(kicad_sch ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not kicad_sch

        Returns:
            - Schematic: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'kicad_sch':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'version': object.version = item[1]
            if item[0] == 'generator': object.generator = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'paper': object.paper = PageSettings().from_sexpr(item)
            if item[0] == 'title_block': object.titleBlock = TitleBlock().from_sexpr(item)
            if item[0] == 'lib_symbols':
                for symbol in item[1:]:
                    object.libSymbols.append(Symbol().from_sexpr(symbol))
            if item[0] == 'junction': object.junctions.append(Junction().from_sexpr(item))
            if item[0] == 'no_connect': object.noConnects.append(NoConnect().from_sexpr(item))
            if item[0] == 'bus_entry': object.busEntries.append(BusEntry().from_sexpr(item))
            if item[0] == 'wire': object.graphicalItems.append(Connection().from_sexpr(item))
            if item[0] == 'bus': object.graphicalItems.append(Connection().from_sexpr(item))
            if item[0] == 'polyline': object.graphicalItems.append(PolyLine().from_sexpr(item))
            if item[0] == 'arc': object.shapes.append(Arc.from_sexpr(item))
            if item[0] == 'circle': object.shapes.append(Circle.from_sexpr(item))
            if item[0] == 'rectangle': object.shapes.append(Rectangle.from_sexpr(item))
            if item[0] == 'image': object.images.append(Image().from_sexpr(item))
            if item[0] == 'text': object.texts.append(Text().from_sexpr(item))
            if item[0] == 'text_box': object.textBoxes.append(TextBox().from_sexpr(item))
            if item[0] == 'label': object.labels.append(LocalLabel().from_sexpr(item))
            if item[0] == 'global_label': object.globalLabels.append(GlobalLabel().from_sexpr(item))
            if item[0] == 'hierarchical_label': object.hierarchicalLabels.append(HierarchicalLabel().from_sexpr(item))
            if item[0] == 'netclass_flag': object.netclassFlags.append(NetclassFlag.from_sexpr(item))
            if item[0] == 'symbol': object.schematicSymbols.append(SchematicSymbol().from_sexpr(item))
            if item[0] == 'sheet': object.sheets.append(HierarchicalSheet().from_sexpr(item))
            if item[0] == 'sheet_instances':
                for instance in item[1:]:
                    object.sheetInstances.append(HierarchicalSheetInstance().from_sexpr(instance))
            if item[0] == 'symbol_instances':
                for instance in item[1:]:
                    object.symbolInstances.append(SymbolInstance().from_sexpr(instance))
        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> Schematic:
        """Load a schematic directly from a KiCad schematic file (`.kicad_sch`) and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - Schematic: Object of the Schematic class initialized with the given KiCad schematic
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            item = cls.from_sexpr(sexpr.parse_sexp(infile.read()))
            item.filePath = filepath
            return item

    @classmethod
    def create_new(cls) -> Schematic:
        """Creates a new empty schematic page with its attributes set as KiCad would create it

        Returns:
            - Schematic: Empty schematic
        """
        schematic = cls(
            version = KIUTILS_CREATE_NEW_VERSION_STR,
            generator = KIUTILS_CREATE_NEW_GENERATOR_STR
        )
        schematic.sheetInstances.append(HierarchicalSheetInstance(instancePath='/', page='1'))
        return schematic

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set, 
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(kicad_sch (version {self.version}) (generator {self.generator})\n'
        if self.uuid is not None:
            expression += f'\n{indents}  (uuid {self.uuid})\n\n'
        expression += f'{self.paper.to_sexpr(indent+2)}'
        if self.titleBlock is not None:
            expression += f'\n{self.titleBlock.to_sexpr(indent+2)}'

        if self.libSymbols:
            expression += f'\n{indents}  (lib_symbols'
            for item in self.libSymbols:
                expression += '\n'
                expression += item.to_sexpr(indent+4)
            expression += f'{indents}  )\n'
        else:
            expression += f'{indents}  (lib_symbols)\n'

        if self.junctions:
            expression += '\n'
            for item in self.junctions:
                expression += item.to_sexpr(indent+2)

        if self.noConnects:
            expression += '\n'
            for item in self.noConnects:
                expression += item.to_sexpr(indent+2)

        if self.busEntries:
            expression += '\n'
            for item in self.busEntries:
                expression += item.to_sexpr(indent+2)

        if self.graphicalItems:
            expression += '\n'
            for item in self.graphicalItems:
                expression += item.to_sexpr(indent+2)

        if self.shapes:
            expression += '\n'
            for item in self.shapes:
                expression += item.to_sexpr(indent+2)

        if self.images:
            expression += '\n'
            for item in self.images:
                expression += item.to_sexpr(indent+2)

        if self.textBoxes:
            expression += '\n'
            for item in self.textBoxes:
                expression += item.to_sexpr(indent+2)

        if self.texts:
            expression += '\n'
            for item in self.texts:
                expression += item.to_sexpr(indent+2)

        if self.labels:
            expression += '\n'
            for item in self.labels:
                expression += item.to_sexpr(indent+2)

        if self.globalLabels:
            expression += '\n'
            for item in self.globalLabels:
                expression += item.to_sexpr(indent+2)

        if self.hierarchicalLabels:
            expression += '\n'
            for item in self.hierarchicalLabels:
                expression += item.to_sexpr(indent+2)

        if self.netclassFlags:
            expression += '\n'
            for item in self.netclassFlags:
                expression += item.to_sexpr(indent+2)

        if self.schematicSymbols:
            for item in self.schematicSymbols:
                expression += '\n'
                expression += item.to_sexpr(indent+2)

        if self.sheets:
            for item in self.sheets:
                expression += '\n'
                expression += item.to_sexpr(indent+2)

        if self.sheetInstances:
            expression += '\n'
            expression += '  (sheet_instances\n'
            for item in self.sheetInstances:
                expression += item.to_sexpr(indent+4)
            expression += '  )\n'

        if self.symbolInstances:
            expression += '\n'
            expression += '  (symbol_instances\n'
            for item in self.symbolInstances:
                expression += item.to_sexpr(indent+4)
            expression += '  )\n'

        expression += f'{indents}){endline}'
        return expression



"""Helper functions for the unittests

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

from dataclasses import dataclass
from typing import Optional
import filecmp
import os

TEST_BASE = os.path.join('tests', 'testdata')

@dataclass
class TestData():
    """Data container to relay testcase-specific information to the report generator. May be added
    as a member to ``unittest.TestCase`` in the ``setUp()`` function. The object will then be
    available in ``result._TestInfo()`` classes constructor."""
    producedOutput: Optional[str] = None
    expectedOutput: Optional[str] = None
    ownDescription: Optional[str] = None
    pathToTestFile: Optional[str] = None
    compareToTestFile: bool = False
    wasSuccessful: bool = False


def to_file_and_compare(object, test_data: TestData) -> bool:
    """Write the object to a file using its ``to_file()`` method and comparing the output with
    the given expected output supplied by a file with ``.expected`` suffix

    Cleans up the test files afterwards, leaving only failing test outputs behind. Test output
    is furthermore saved to ``test_data.expectedOutput`` and ``test_data.producedOutput`` to be
    displayed in the HTML test report.

    Args:
        - object: KiUtils object with a ``to_file()`` method
        - test_data (TestData): Test data object of the currently running test (contains path to 
                                test file)

    Returns:
        - bool: True, if both the output of ``to_file()`` and the given expected output are the same
    """
    # Create S-Expression from object
    if test_data.pathToTestFile is None:
        raise Exception("pathToTestFile may not be None!")

    object.to_file(f'{test_data.pathToTestFile}.testoutput')

    # Compare with the expected result
    if test_data.compareToTestFile:
        compare_file = test_data.pathToTestFile
    else:
        compare_file = f'{test_data.pathToTestFile}.expected'

    test_data.wasSuccessful = filecmp.cmp(f'{test_data.pathToTestFile}.testoutput', compare_file)
    cleanup_after_test(test_data)
    return test_data.wasSuccessful

def load_contents(file: str) -> str:
    """Load contents of a specific file and return it as a joined string

    Args:
        file (str): Path to file

    Returns:
        str: Contents of file in one string
    """
    with open(file, "r") as outfile:
        return ''.join(outfile.readlines())

def prepare_test(object):
    """Prepare a unittest test case in the KiUtils framework

    Args:
        object: Test case class object
    """
    object.testData = TestData()

def cleanup_after_test(test_data: TestData):
    """Cleanup after a unittest test case ran

    Args:
        test_data (TestData): Test data structure of test case that finished last

    Raises:
        Exception: When ``test_data.pathToTestFile`` is None
    """
    if test_data.pathToTestFile is None:
        raise Exception("Path to testfile must not be None!")
    test_data.producedOutput = load_contents(f'{test_data.pathToTestFile}.testoutput')
    if test_data.compareToTestFile:
        test_data.expectedOutput = load_contents(f'{test_data.pathToTestFile}')
    else:
        test_data.expectedOutput = load_contents(f'{test_data.pathToTestFile}.expected')

    if test_data.wasSuccessful:
        os.remove(f'{test_data.pathToTestFile}.testoutput')


"""Footprint graphical items define all of the drawing items that are used in the footprint
definition. This includes text, text boxes, lines, rectangles, circles, arcs, polygons, curves
and dimensions.

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    08.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_graphics_items
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List

from kiutils.items.common import RenderCache, Stroke, Position, Effects
from kiutils.utils.strings import dequote

# FIXME: Several classes have a ``stroke`` member. This feature will be introduced in KiCad 7 and
#        has yet to be tested here.

@dataclass
class FpText():
    """The ``fp_text`` token defines a graphic line in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_text
    """

    type: str = "reference"
    """The ``type`` attribute defines the type of text. Valid types are ``reference``, ``value``, and
    ``user``"""

    text: str = "%REF"
    """The ``text`` attribute is a string that defines the text"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y position coordinates and optional orientation angle of
    the text"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the text resides on"""

    hide: bool = False
    """The optional ``hide`` token, defines if the text is hidden"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is displayed"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the text object"""

    renderCache: Optional[RenderCache] = None
    """If the ``effects`` token prescribe a TrueType font then the optional ``render_cache`` token 
    should be given in case the font can not be found on the current system.
    
    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpText:
        """Convert the given S-Expresstion into a FpText object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_text ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_text

        Returns:
            - FpText: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_text':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.type = exp[1]
        object.text = exp[2]
        for item in exp[3:]:
            if type(item) != type([]):
                if item == 'hide': object.hide = True
                continue
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'render_cache': object.renderCache = RenderCache.from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        hide = ' hide' if self.hide else ''
        unlocked = ' unlocked' if self.position.unlocked else ''
        posA = f' {self.position.angle}' if self.position.angle is not None else ''

        expression =  f'{indents}(fp_text {self.type} "{dequote(self.text)}" (at {self.position.X} {self.position.Y}{posA}{unlocked}) (layer "{dequote(self.layer)}"){hide}\n'
        expression += f'{indents}  {self.effects.to_sexpr()}'
        if self.tstamp is not None:
            expression += f'{indents}  (tstamp {self.tstamp})\n'
        if self.renderCache is not None:
            expression += self.renderCache.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class FpLine():
    """The ``fp_line`` token defines a graphic line in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_line
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the upper left corner of the line"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the low right corner of the line"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the line resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the line. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the line. (version 7)"""

    # FIXME: This is not implemented in to_sexpr() because it does not seem to be used on lines
    #        in footprints. Further testing required ..
    locked: bool = False
    """The optional ``locked`` token defines if the line cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the line object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpLine:
        """Convert the given S-Expresstion into a FpLine object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_line ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_line

        Returns:
            - FpLine: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_line':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        return f'{indents}(fp_line (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y}) (layer "{dequote(self.layer)}"){width}{tstamp}){endline}'

@dataclass
class FpRect():
    """The ``fp_rect`` token defines a graphic rectangle in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_rectangle
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the upper left corner of the rectangle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the low right corner of the rectangle"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the rectangle resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the rectangle. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the rectangle. (version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the rectangle is filled. Valid fill types are solid
    and none. If not defined, the rectangle is not filled."""

    locked: bool = False
    """The optional ``locked`` token defines if the rectangle cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the rectangle object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpRect:
        """Convert the given S-Expresstion into a FpRect object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_rect ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_rect

        Returns:
            - FpRect: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_rect':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = ' locked' if self.locked else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''

        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        return f'{indents}(fp_rect (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y}) (layer "{dequote(self.layer)}"){width}{fill}{locked}{tstamp}){endline}'

@dataclass
class FpTextBox():
    """The ``fp_text_box`` token defines a rectangle containing line-wrapped text.
    
    Available since KiCad v7
    
    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_text_box
    """

    locked: bool = False
    """The ``locked`` token specifies if the text box can be moved. Defaults to ``False``."""

    text: str = ""
    """The ``text`` token defines the content of the text box. Defaults to an empty string."""

    start: Optional[Position] = None
    """The optional ``start`` token defines the top-left of a cardinally oriented text box"""

    end: Optional[Position] = None
    """The optional ``end`` token defines the bottom-right of a cardinally oriented text box"""

    pts: List[Position] = field(default_factory=list)
    """The ``pts`` token defines the four corners of a non-cardianlly oriented text box. The corners
    must be in order, but the winding can be either direction."""

    angle: Optional[float] = None
    """The optional ``angle`` token defines the rotation of the text box in degrees.

    Note:
        - If ``angle`` is not given, or is a cardinal angle (0, 90, 180 or 270), then the text box
        MUST have ``start`` and ``end`` tokens.
        - If ``angle`` is given and is not a cardinal angle, then the text box MUST have a ``pts``
        token (with 4 pts).
    """

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the text box resides on. Defaults to
    ``F.Cu``."""

    tstamp: Optional[str] = None
    """The optional ``tstamp`` token defines the unique identifier of the text box"""

    effects: Optional[Effects] = None
    """The optional ``effects`` token describes the style of the text in the text box"""

    stroke: Optional[Stroke] = None
    """The optional ``stroke`` token describes the style of an optional border to be drawn around 
    the text box"""

    renderCache: Optional[RenderCache] = None
    """If the ``effects`` token prescribe a TrueType font then the optional ``render_cache`` token 
    should be given in case the font can not be found on the current system."""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpTextBox:
        """Convert the given S-Expression into a FpTextBox object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_text_box ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_text_box

        Returns:
            - FpTextBox: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_text_box':
            raise Exception("Expression does not have the correct type")

        object = cls()

        # Extract "locked" token, if any is present
        if exp[1] == "locked" and not isinstance(exp[2], list):
            object.locked = True
            object.text = exp[2]
            start_at = 3
        else:
            object.text = exp[1]
            start_at = 2

        for item in exp[start_at:]:
            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'pts':
                for point in item[1:]:
                    object.pts.append(Position().from_sexpr(point))
            if item[0] == 'angle': object.angle = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'effects': object.effects = Effects.from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke.from_sexpr(item)
            if item[0] == 'render_cache': object.renderCache = RenderCache.from_sexpr(item)

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Raises:
            - Exception: When a non-cardinal angle is given and no corner points were defined using
              the ``self.pts`` token
            - Exception: When a cardinal angle or no angle is given and either start or end token
              is undefined

        Returns:
            - str: S-Expression of this object
        """
        if self.angle is not None and self.angle not in [0.0, 90.0, 180.0, 270.0]:
            if len(self.pts) != 4:
                raise Exception("None-cardinal angles must have exactly four corner points defined")
        if self.angle is None or self.angle in [0.0, 90.0, 180.0, 270.0]:
            if self.start is None or self.end is None:
                raise Exception("No angle or a cardinal angle needs a start and end token defined")

        indents = ' '*indent
        endline = '\n' if newline else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        angle = f'(angle {self.angle}) ' if self.angle is not None else ''
        start = f'(start {self.start.X} {self.start.Y}) ' if self.start is not None else ''
        end = f'(end {self.end.X} {self.end.Y}) ' if self.end is not None else ''
        locked = ' locked' if self.locked else ''

        expression = f'{indents}(fp_text_box{locked} "{dequote(self.text)}"\n'
        if len(self.pts) == 4:
            expression += f'{indents}  (pts\n'
            expression += f'{indents}    (xy {self.pts[0].X} {self.pts[0].Y})      (xy {self.pts[1].X} {self.pts[1].Y})      (xy {self.pts[2].X} {self.pts[2].Y})      (xy {self.pts[3].X} {self.pts[3].Y})\n'
            expression += f'{indents}  )\n'
        expression += f'{indents}  {start}{end}{angle}(layer "{dequote(self.layer)}"){tstamp}\n'
        if self.effects is not None:
            expression += self.effects.to_sexpr(indent+2)
        if self.stroke is not None:
            expression += self.stroke.to_sexpr(indent+2)
        if self.renderCache is not None:
            expression += self.renderCache.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class FpCircle():
    """The ``fp_circle `` token defines a graphic circle in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_circle
    """

    center: Position = field(default_factory=lambda: Position())
    """The ``center`` token defines the coordinates of the center of the circle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the low right corner of the circle"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the circle resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the circle. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the circle. (version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the circle is filled. Valid fill types are ``solid``
    and ``none``. If not defined, the circle is not filled."""

    locked: bool = False
    """The optional ``locked`` token defines if the circle cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the circle object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpCircle:
        """Convert the given S-Expresstion into a FpCircle object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_circle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_circle

        Returns:
            - FpCircle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_circle':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'center': object.center = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = ' locked' if self.locked else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''

        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        return f'{indents}(fp_circle (center {self.center.X} {self.center.Y}) (end {self.end.X} {self.end.Y}) (layer "{dequote(self.layer)}"){width}{fill}{locked}{tstamp}){endline}'

@dataclass
class FpArc():
    """The ``fp_arc`` token defines a graphic arc in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_arc
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the start position of the arc radius"""

    mid: Position = field(default_factory=lambda: Position())
    """The ``mid`` token defines the coordinates of the midpoint along the arc"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the end position of the arc radius"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the arc resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the arc. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the arc. (version 7)"""

    locked: bool = False
    """The optional ``locked`` token defines if the arc cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the arc object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpArc:
        """Convert the given S-Expresstion into a FpArc object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_arc ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_arc

        Returns:
            - FpArc: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_arc':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'mid': object.mid = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = ' locked' if self.locked else ''

        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        return f'{indents}(fp_arc (start {self.start.X} {self.start.Y}) (mid {self.mid.X} {self.mid.Y}) (end {self.end.X} {self.end.Y}) (layer "{dequote(self.layer)}"){width}{locked}{tstamp}){endline}'

@dataclass
class FpPoly():
    """The ``fp_poly`` token defines a graphic polygon in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_polygon
    """

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the polygon resides on"""

    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` define the list of X/Y coordinates of the polygon outline"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the polygon. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the polygon. (version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the polygon is filled. Valid fill types are solid
    and none. If not defined, the rectangle is not filled."""

    locked: bool = False
    """The optional ``locked`` token defines if the polygon cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the polygon object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpPoly:
        """Convert the given S-Expresstion into a FpPoly object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_poly ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_poly

        Returns:
            - FpPoly: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_poly':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'pts':
                for point in item[1:]:
                    object.coordinates.append(Position().from_sexpr(point))
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the polygon, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = ' locked' if self.locked else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''

        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        expression = f'{indents}(fp_poly (pts\n'
        for point in self.coordinates:
            expression += f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  ) (layer "{dequote(self.layer)}"){width}{fill}{locked}{tstamp}){endline}'
        return expression

@dataclass
class FpCurve():
    """The ``fp_curve`` token defines a graphic Cubic Bezier curve in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_curve
    """

    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` define the list of X/Y coordinates of the curve outline"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the curve resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the curve. (prior to version 7)"""

    stroke: Optional[Stroke] = None   # Used for KiCad >= 7
    """The ``stroke`` describes the line width and style of the curve. (version 7)"""

    locked: bool = False
    """The optional ``locked`` token defines if the curve cannot be edited"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the curve object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FpCurve:
        """Convert the given S-Expresstion into a FpCurve object

        Args:
            - exp (list): Part of parsed S-Expression ``(fp_curve ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_curve

        Returns:
            - FpCurve: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fp_curve':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'pts':
                for point in item[1:]:
                    object.coordinates.append(Position().from_sexpr(point))
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width':
                object.width = item[1]
                object.stroke = None
            if item[0] == 'stroke':
                object.stroke = Stroke.from_sexpr(item)
                object.width = None

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the curve, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = ' locked' if self.locked else ''

        if self.width is not None:
            width = f' (width {self.width})'
        elif self.stroke is not None:
            width = f' {self.stroke.to_sexpr(indent=0, newline=False)}'
        else:
            width = ''

        expression = f'{indents}(fp_curve (pts\n'
        for point in self.coordinates:
            expression += f'{indents}  (xy {point.X} {point.Y})\n'
        expression += f'{indents}) (layer "{dequote(self.layer)}"){width}{locked}{tstamp}){endline}'
        return expression



"""Classes to manage KiCad footprint and symbol library tables

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    19.02.2022 - created
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List
from os import path

from kiutils.utils.strings import dequote
from kiutils.utils import sexpr

@dataclass
class Library():
    """The ``library`` token defines either a symbol library or a footprint library in
    a library table file (``fp_lib_table`` or ``sym_lib_table``)"""

    name: str = ""
    """The ``name`` token defines the name of the library as displayed in the project"""

    type: str = "KiCad"
    """The ``type`` token defines the type of the library, usually ``KiCad``"""

    uri: str = ""
    """The ``uri`` token defines the path to the library files"""

    options: str = ""
    """The ``options`` token (..) TBD"""

    description: str = ""
    """The ``description`` token (..) TBD"""

    active: bool = True
    """The ``active`` token sets if the library is loaded by KiCad"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Library:
        """Convert the given S-Expresstion into a Library object

        Args:
            - exp (list): Part of parsed S-Expression ``(lib ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not lib

        Returns:
            - Library: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'lib':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'type': object.type = item[1]
            if item[0] == 'uri': object.uri = item[1]
            if item[0] == 'options': object.options = item[1]
            if item[0] == 'descr': object.description = item[1]
            if item[0] == 'disabled': object.active = False
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression = f'{indents}(lib '
        expression += f'(name "{dequote(self.name)}")'
        expression += f'(type "{dequote(self.type)}")'
        expression += f'(uri "{dequote(self.uri)}")'
        expression += f'(options "{dequote(self.options)}")'
        expression += f'(descr "{dequote(self.description)}")'

        if not self.active:
            expression += '(disabled)'

        expression += f'){endline}'

        return expression

@dataclass
class LibTable():
    """The ``libtable`` token defines the ``fp_lib_table`` or ``sym_lib_table`` file of KiCad"""

    type: str = 'sym_lib_table'
    """The ``type`` token defines the type of the library table. Valid values are ``fp_lib_table`` or
    ``sym_lib_table``."""

    libs: List[Library] = field(default_factory=list)
    """The ``libs`` token holds a list of librarys that this library table object holds"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the library file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> LibTable:
        """Convert the given S-Expresstion into a LibTable object

        Args:
            - exp (list): Part of parsed S-Expression ``(sym_lib_table ...)`` or ``(fp_lib_table ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not lib

        Returns:
            - LibTable: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if not (exp[0] == 'fp_lib_table' or exp[0] == 'sym_lib_table'):
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.type = exp[0]
        for item in exp:
            if item[0] == 'lib': object.libs.append(Library().from_sexpr(item))
        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> LibTable:
        """Load a library table directly from a KiCad library table file and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - LibTable: Object of the LibTable class initialized with the given KiCad library table
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            item = cls.from_sexpr(sexpr.parse_sexp(infile.read()))
            item.filePath = filepath
            return item

    @classmethod
    def create_new(cls, type: str = 'sym_lib_table') -> LibTable:
        """Creates a new empty library table with its attributes set as KiCad would create it

        Args:
            - type (str): ``fp_lib_table`` or ``sym_lib_table``. Defaults to the latter.

        Returns:
            - Library: Empty library table of given type
        """
        return cls(type=type)

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set, 
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression = f'{indents}({self.type}\n'
        for lib in self.libs:
            expression += lib.to_sexpr()
        expression += f'{indents}){endline}'
        return expression


"""Functions for string manipulation

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    28.02.2022 - created
"""

def dequote(input: str) -> str:
    """Escapes double-quotes in a string using a backslash

    Args:
        - input (str): String to replace double-quotes

    Returns:
        - str: String with replaced double-quotes
    """
    return str(input).replace("\"", "\\\"")


def remove_prefix(input: str, prefix: str) -> str:
    """Removes the given prefix from a string (to remove incompatibility of ``str.removeprefix()``
    for Python versions < 3.9)

    Args:
        - input (str): String to remove the prefix from
        - prefix (str): The prefix

    Returns:
        - str: String with removed prefix, or the ``input`` string as is, if the prefix was not found
    """
    return input[len(prefix):] if input.startswith(prefix) else input